
#include "application/camera_control.hpp"

#include "math/vector.hpp"
#include <exception>

namespace _462 {

#define CAMERA_TRANSLATION_SCALE_FACTOR 0.01
#define CAMERA_ROTATION_SCALE_FACTOR 0.01

/* different camera operations performed by gui */
enum CameraAction
{
    CA_NONE, // represents event of doing nothing
    CA_FOCUS_PITCH,
    CA_FOCUS_YAW,
    CA_FOCUS_ZOOM,
    CA_TRANSLATE_X,
    CA_TRANSLATE_Y,
    CA_TRANSLATE_Z,
    CA_ROLL,
    CA_PITCH,
    CA_YAW
};

struct MouseState
{
    // mouse position as of last callback
    int x,y;
    // button state of the left, middle, and right buttons. true means
    // down, false means up
    bool button_state[3];

    /* The current camera control state of the x and y screen axes. */
    CameraAction camera_control[2];

    MouseState():
        x(0), y(0)
    {
        button_state[Application::MOUSE_LEFT] = false;
        button_state[Application::MOUSE_MIDDLE] = false;
        button_state[Application::MOUSE_RIGHT] = false;
        camera_control[0] = CA_NONE;
        camera_control[1] = CA_NONE;
    }
};

enum {
    MODIFIER_INDEX_NONE = 0,
    MODIFIER_INDEX_CTRL,
    MODIFIER_INDEX_SHIFT
};

/*
    Camera controls:
    left-click, x-axis yaws around focus
    left-click, y-axis pitches around focus
    middle-click, y-axis zooms toward/away from focus

    ctrl + left-click, x-axis controls translation along local x-axis
    ctrl + left-click, y-axis controls translation along local y-axis
    ctrl + middle-click, y-axis controls translation along local z-axis

    shift + left-click, x-axis controls yaw
    shift + left-click, y-axis controls pitch
    shift + middle-click, y-axis controls roll

    array elements are indexed by
    control_map[modifier key state][button][mouse axis]
 */
static const CameraAction control_map[3][2][2] = {
    {
        { CA_FOCUS_YAW  , CA_FOCUS_PITCH },
        { CA_NONE       , CA_FOCUS_ZOOM  }
    },
    {
        { CA_TRANSLATE_X, CA_TRANSLATE_Y },
        { CA_NONE       , CA_TRANSLATE_Z }
    },
    {
        { CA_YAW        , CA_PITCH       },
        { CA_NONE       , CA_ROLL,       }
    }
};

class CameraControl::CameraControlImpl
{
public:

    void apply_control( Camera& camera, CameraAction ca, int delta );

    MouseState mouse;

};

CameraControl::CameraControl()
{
    impl = new CameraControlImpl();
    if ( !impl )
        throw 0;
}

CameraControl::~CameraControl()
{
    assert( impl );
    delete impl;
}

void CameraControl::on_mouse_press( Window* window, Application::MouseButton button, int x, int y, size_t modifiers, bool pressed )
{
    MouseState& mouse = impl->mouse;

    mouse.button_state[button] = pressed;

    // left and middle button control camera
    if ( button != Application::MOUSE_RIGHT ) {
        // if either is released, cancel controls
        if ( !pressed ) {
            mouse.camera_control[0] = CA_NONE;
            mouse.camera_control[1] = CA_NONE;
        } else {
            int modifier_index;

            if ( modifiers & Application::MODIFIER_CTRL )
                modifier_index = MODIFIER_INDEX_CTRL;
            else if ( modifiers & Application::MODIFIER_SHIFT )
                modifier_index = MODIFIER_INDEX_SHIFT;
            else
                modifier_index = MODIFIER_INDEX_NONE;

            mouse.camera_control[0] = control_map[modifier_index][button][0];
            mouse.camera_control[1] = control_map[modifier_index][button][1];
        }
    }
}

void CameraControl::on_mouse_move( Window* window, int x, int y )
{
    MouseState& mouse = impl->mouse;

    int oldx = mouse.x;
    int oldy = mouse.y;

    mouse.x = x;
    mouse.y = y;

    if ( x != oldx )
        impl->apply_control( camera, mouse.camera_control[0], x - oldx );
    if ( y != oldy )
        impl->apply_control( camera, mouse.camera_control[1], y - oldy );
}

/**
 * Applies the given control to the current scene's camera.
 * @param ca The control to apply.
 * @param delta The distance of mouse movement.
 */
void CameraControl::CameraControlImpl::apply_control( Camera& camera, CameraAction ca, int delta )
{
    // if nothing, just return
    if ( ca == CA_NONE )
        return;

    real_t angle = CAMERA_ROTATION_SCALE_FACTOR * delta;
    real_t distance = CAMERA_TRANSLATION_SCALE_FACTOR * delta;
    real_t new_focus_dist;

    switch ( ca )
    {
    case CA_FOCUS_YAW:
        camera.yaw_about_focus( -angle );
        break;

    case CA_FOCUS_PITCH:
        camera.pitch_about_focus( -angle );
        break;

    case CA_FOCUS_ZOOM:
        // change focus based on an exponential scale
        new_focus_dist = pow( 2, distance ) * camera.focus_dist;
        camera.translate( Vector3::UnitZ * ( new_focus_dist - camera.focus_dist ) );
        camera.focus_dist = new_focus_dist;
        break;

    case CA_TRANSLATE_X:
        camera.translate( -Vector3::UnitX * distance );
        break;

    case CA_TRANSLATE_Y:
        camera.translate( Vector3::UnitY * distance );
        break;

    case CA_TRANSLATE_Z:
        camera.translate( Vector3::UnitZ * distance );
        break;

    case CA_ROLL:
        camera.roll( angle );
        break;

    case CA_PITCH:
        camera.pitch( angle );
        break;

    case CA_YAW:
        camera.yaw( angle );
        break;

    default:
        break;
    }
}

} /* _462 */

