/**
 * @file main2d.cpp
 * @brief The main function and application for p1 2d part.
 *
 * @author Eric Butler (edbutler)
 */

#include "application/application.hpp"

#include "opengl.hpp"
#include "p1_2d.hpp"
#include "math/vector.hpp"
#include <iostream>

namespace _462 {

// contstants for 2d spline

// the pixel size for control points
static const real_t S2D_CP_PIXEL_SIZE = 5.0;
// the color of control points
static const float S2D_CP_COLOR[3] = { 0.1, 0.5, 0.7 };
// the color of the spline
static const float S2D_SP_COLOR[3] = { 0.8, 0.45, 0.05 };

class Spline2DApplication : public Application
{
public:

    Spline2DApplication() { }
    virtual ~Spline2DApplication() { }

    virtual bool initialize( Window* window, int argc, char* argv[] );
    virtual void destroy( Window* ) { }
    virtual void update( Window*, real_t ) { }
    virtual void render( Window* window );

    virtual void on_mouse_press( Window* window, MouseButton button, int x, int y, size_t modifiers, bool pressed );

    void add_new_2d_control_point( Window* window, int x, int y );
    void render_2d_spline( Window* window );

private:

    static void hermite_callback( Application* application, Window* )
    {
        Spline2DApplication* app = dynamic_cast<Spline2DApplication*>( application );
        if ( app ) {
            app->spline_type = ST_HERMITE;
        }
    }

    static void catmull_callback( Application* application, Window* )
    {
        Spline2DApplication* app = dynamic_cast<Spline2DApplication*>( application );
        if ( app ) {
            app->spline_type = ST_CATMULL_ROM;
        }
    }

    static void bezier_callback( Application* application, Window* )
    {
        Spline2DApplication* app = dynamic_cast<Spline2DApplication*>( application );
        if ( app ) {
            app->spline_type = ST_BEZIER;
        }
    }

    ControlPoints control_points;
    SplineType spline_type;
};

static void print_usage( const char* prog_name )
{
    std::cout << "Usage: " << prog_name << " {hermite|catmull-rom|bezier}" << std::endl;
}

bool Spline2DApplication::initialize( Window* window, int argc, char* argv[] )
{
    static const int TYPE_ARG_INDEX = 1;

    // need arg for spline type
    if ( argc < TYPE_ARG_INDEX + 1 ) {
        print_usage( argv[0] );
        return false;
    }

    std::string spline_type_str = argv[TYPE_ARG_INDEX];

    static const std::string ST_HERMITE_STR = "hermite";
    static const std::string ST_CATMULL_ROM_STR = "catmull-rom";
    static const std::string ST_BEZIER_STR = "bezier";

    // parse the spline type
    if ( spline_type_str == ST_HERMITE_STR ) {
        spline_type = ST_HERMITE;
    } else if ( spline_type_str == ST_CATMULL_ROM_STR ) {
        spline_type = ST_CATMULL_ROM;
    } else if ( spline_type_str == ST_BEZIER_STR ) {
        spline_type = ST_BEZIER;
    } else {
        std::cout << "Unkown spline type" << std::endl;
        print_usage( argv[0] );
        return false;
    }

    window->add_menu_item( "Hermite", hermite_callback );
    window->add_menu_item( "Catmull-Rom", catmull_callback );
    window->add_menu_item( "Bezier", bezier_callback );

    return true;
}

void Spline2DApplication::render( Window* window )
{
    int width, height;

    // query current window size
    window->get_dimension( &width, &height );

    // resize viewport
    glViewport( 0, 0, width, height );

    // set matrices
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluOrtho2D( 0, width, 0, height );

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();

    // clear buffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // render the spline
    glColor3fv( S2D_SP_COLOR );
    prj_2d_render( control_points, spline_type );

    // render the control points
    glPointSize( S2D_CP_PIXEL_SIZE );
    glColor3fv( S2D_CP_COLOR );

    glBegin( GL_POINTS );
    for ( ControlPoints::const_iterator i = control_points.begin(); i != control_points.end(); ++i ) {
        Vector3 point( *i, 0 );
        glVertex3d( point.x, point.y, point.z );
    }
    glEnd();
}

void Spline2DApplication::on_mouse_press( Window* window, MouseButton button, int x, int y, size_t modifiers, bool pressed )
{
    if ( pressed ) {
        int width, height;
        window->get_dimension( &width, &height );
        // add the control point, compensating for inverted y
        control_points.push_back( Vector2( x, height - y ) );
    }
}


} /* _462 */

using namespace _462;

/**
 * Application entry point.
 */
int main( int argc, char* argv[] )
{
    int width = 800;
    int height = 600;
    real_t fps = 30.0;

    const char* title = "462 Project 1 - 2D Splines";
    Window::start_application( new _462::Spline2DApplication(),
                               argc, argv, width, height, fps, title );

    return 0;
}

