/**
 * @file maincoaster.cpp
 * @brief The main function and application for p1 coaster part.
 *
 * @author Eric Butler (edbutler)
 */

#include "application/application.hpp"

#include "application/camera_control.hpp"
#include "application/spline_parser.hpp"

#include "opengl.hpp"
#include "p1_coaster.hpp"
#include "math/vector.hpp"
#include <iostream>

namespace _462 {

class SplineCoasterApplication : public Application
{
public:

    SplineCoasterApplication() { }
    virtual ~SplineCoasterApplication() { }

    virtual bool initialize( Window* window, int argc, char* argv[] );
    virtual void destroy( Window* ) { }
    virtual void update( Window* window, real_t time );
    virtual void render( Window* window );

    virtual void on_mouse_press( Window* window, MouseButton button, int x, int y, size_t modifiers, bool pressed );
    virtual void on_mouse_move( Window* window, int x, int y );

private:

    // camera control. not really needed for this part, but useful for debugging.
    CameraControl camera_control;
};

static void print_usage( const char* prog_name )
{
    std::cout << "Usage: " << prog_name
        << " <control point file>" << std::endl;
}

void SplineCoasterApplication::on_mouse_press( Window* window, MouseButton button, int x, int y, size_t modifiers, bool pressed )
{
    camera_control.on_mouse_press( window, button, x, y, modifiers, pressed );
}

void SplineCoasterApplication::on_mouse_move( Window* window, int x, int y )
{
    camera_control.on_mouse_move( window, x, y );
}

bool SplineCoasterApplication::initialize( Window* window, int argc, char* argv[] )
{
    // need second arg for model
    if ( argc < 2 ) {
        print_usage( argv[0] );
        return false;
    }

    Camera* camera = &camera_control.camera;

    // set camera paremeters
    camera->fov = PI / 4.0;
    camera->near_clip = 0.1;
    camera->far_clip = 1000.0;
    // aspect is set by the application, so we can ignore it

    camera->position = Vector3( 0.0, 0.0, 5.0 );
    camera->orientation = Quaternion::Identity;
    // set focus point at the origin
    camera->focus_dist = camera->position.distance( Vector3::Zero );

    // change the pixel alignment requirements
    glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
    glPixelStorei( GL_PACK_ALIGNMENT, 1 );

    // parse the spline file

    CatmullRomControlPointList list;
    real_t tau;

    if ( !parse_spline_file( argv[1], list, &tau )) {
        std::cout << "Error parsing spline file\n";
        return false;
    }

    p1_coaster_initialize( list, tau );

    return true;
}

void SplineCoasterApplication::update( Window*, real_t delta_time )
{
    p1_coaster_update( delta_time );
}

void SplineCoasterApplication::render( Window* window )
{
    int width, height;

    // query current window size, resize viewport
    window->get_dimension( &width, &height );
    glViewport( 0, 0, width, height );

    camera_control.camera.aspect = real_t( width ) / real_t( height );

    p1_coaster_render( camera_control.camera );
}

} /* _462 */

using namespace _462;

/**
 * Application entry point.
 */
int main( int argc, char* argv[] )
{
    int width = 800;
    int height = 600;
    real_t fps = 30.0;
    const char* title = "462 Project 1 - Roller Coaster";

    // start a new application
    Window::start_application( new _462::SplineCoasterApplication(),
                               argc, argv, width, height, fps, title );

    return 0;
}

