/**
 * @file mainmodel.cpp
 * @brief The main function and application for p1 model part.
 *
 * @author Eric Butler (edbutler)
 */

#include "application/application.hpp"
#include "application/camera_control.hpp"

#include "opengl.hpp"
#include "p1_model.hpp"
#include "math/vector.hpp"
#include "scene/model.hpp"
#include <iostream>

namespace _462 {

class SplineModelApplication : public Application
{
public:

    SplineModelApplication() { }
    virtual ~SplineModelApplication() { }

    virtual bool initialize( Window* window, int argc, char* argv[] );
    virtual void destroy( Window* ) { }
    virtual void update( Window* window, real_t time );
    virtual void render( Window* window );

    virtual void on_mouse_press( Window* window, MouseButton button, int x, int y, size_t modifiers, bool pressed );
    virtual void on_mouse_move( Window* window, int x, int y );

private:

    CameraControl camera_control;

    Model model;
};

static void print_usage( const char* prog_name )
{
    std::cout << "Usage: " << prog_name
        << " <obj file> <texture file> [animaiton name]" << std::endl;
}

bool SplineModelApplication::initialize( Window* window, int argc, char* argv[] )
{
    // need second arg for model
    if ( argc < 3 ) {
        print_usage( argv[0] );
        return false;
    }

    // load model
    if ( !model.load_from_obj( argv[1] ) ) {
        std::cout << "Error loading model. Aborting." << std::endl;
        print_usage( argv[0] );
        return false;
    }

    model.texture_filename = argv[2];

    Camera* camera = &camera_control.camera;

    // set camera paremeters
    camera->fov = PI / 4.0;
    camera->near_clip = 0.1;
    camera->far_clip = 100.0;
    // aspect is set by the application, so we can ignore it

    camera->position = Vector3( 0.0, 0.0, 5.0 );
    camera->orientation = Quaternion::Identity;
    // set focus point at the origin
    camera->focus_dist = camera->position.distance( Vector3::Zero );

    // change the pixel alignment requirements
    glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
    glPixelStorei( GL_PACK_ALIGNMENT, 1 );

    const char* animation = argc >= 4 ? argv[3] : "default";
    p1_model_initialize( &model, animation );

    return true;
}

void SplineModelApplication::update( Window*, real_t delta_time )
{
    p1_model_update( delta_time );
}

void SplineModelApplication::render( Window* window )
{
    int width, height;

    // query current window size, resize viewport
    window->get_dimension( &width, &height );
    glViewport( 0, 0, width, height );

    camera_control.camera.aspect = real_t( width ) / real_t( height );

    p1_model_render( camera_control.camera );
}

void SplineModelApplication::on_mouse_press( Window* window, MouseButton button, int x, int y, size_t modifiers, bool pressed )
{
    camera_control.on_mouse_press( window, button, x, y, modifiers, pressed );
}

void SplineModelApplication::on_mouse_move( Window* window, int x, int y )
{
    camera_control.on_mouse_move( window, x, y );
}

} /* _462 */

using namespace _462;

/**
 * Application entry point.
 */
int main( int argc, char* argv[] )
{
    int width = 800;
    int height = 600;
    real_t fps = 30.0;
    const char* title = "462 Project 1 - Model Rendering";

    // start a new application
    Window::start_application( new _462::SplineModelApplication(),
                               argc, argv, width, height, fps, title );

    return 0;
}

